#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a time dependent
    analytical function.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Fri Aug 9 2019

"""

import numpy as np
import pandas as pd
import sys
import joblib
import matplotlib

matplotlib.use('agg')

# import bayesvalidrox
# Add BayesValidRox path
sys.path.append("../../src/")
print(sys.path)

from bayesvalidrox import PyLinkForwardModel, Input, ExpDesigns, GPESkl, PostProcessing, BayesInference, Discrepancy, Engine

if __name__ == "__main__":

    # Number of parameters
    ndim = 2  # 2, 10

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    Model.link_type = 'Function'
    Model.py_file = 'analytical_function'
    Model.name = 'AnalyticFunc'

    Model.Output.names = ['Z']

    # For Bayesian inversion synthetic data with X=[0,0]
    Model.observations = {}
    Model.observations['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.observations['Z'] = np.repeat([2.], 10)

    # For Checking with the MonteCarlo refrence
    Model.mc_reference = {}
    Model.mc_reference['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.mc_reference['mean'] = np.load(f"data/mean_{ndim}.npy")
    Model.mc_reference['std'] = np.load(f"data/std_{ndim}.npy")

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    # Assuming dependent input variables
    # Inputs.Rosenblatt = True

    for i in range(ndim):
        Inputs.add_marginals()
        Inputs.Marginals[i].name = "$\\theta_{" + str(i + 1) + "}$"
        Inputs.Marginals[i].dist_type = 'uniform'
        Inputs.Marginals[i].parameters = [-5, 5]

    # arbitrary polynomial chaos
    # inputParams = np.load('data/InputParameters_{}.npy'.format(ndim))
    # for i in range(ndim):
    #     Inputs.add_marginals()
    #     Inputs.Marginals[i].name = f'$X_{i+1}$'
    #     Inputs.Marginals[i].input_data = inputParams[:, i]

    # =====================================================
    # ==========  DEFINITION OF THE METAMODEL  ============
    # =====================================================
    MetaModelOpts = GPESkl(Inputs)  # , Model)

    # Select if you want to preserve the spatial/temporal depencencies
    # MetaModelOpts.dim_red_method = 'PCA'
    # MetaModelOpts.var_pca_threshold = 99.999
    # MetaModelOpts.n_pca_components = 10

    # ToDo: This should not be an option for GPEs
    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) aPCE (arbitrary PCE)
    # 3) GPE (Gaussian Process Emulator)
    MetaModelOpts.meta_model_type = 'GPE'

    # ------------------------------------------------
    # ------------- GPE Specification ----------------
    # ------------------------------------------------
    # Select the solver for solving for the GP Hyperparameters using the ML approach
    # ToDo: Remove this as a user-defined parameter, since only one is available?
    # 1)LBFGS: only option for Scikit Learn
    MetaModelOpts._gpe_reg_method = 'LBFGS'

    # Kernel options ----------------------------
    # Loop over different Kernels:
    # 1) True to loop over the different kernel types and select the best one
    MetaModelOpts._autoSelect = False

    # Select Kernel type:
    # 1) RBF: Gaussian/squared exponential kernel
    # 2) Matern Kernel
    # 3) RQ: Rational Quadratic kernel
    MetaModelOpts._kernel_type = 'RBF'

    MetaModelOpts.normalize_x_method = 'norm'  # Input data transformation
    MetaModelOpts._kernel_isotropy = True      # Kernel isotropy, False for anisotropy
    MetaModelOpts._nugget = 1e-6                # Set regularization parameter (constant)
    MetaModelOpts._kernel_noise = False         # Optimize regularization parameter. True to consider WhiteKernel
    MetaModelOpts.n_restarts = 20
    # End GP-specific input
    # -------------------------------------------------------------

    # Bootstraping
    # 1) normal 2) fast
    MetaModelOpts.n_bootstrap_itrs = 1

    # Print summary of the regression results
    # MetaModelOpts.verbose = True

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    ExpDesign = ExpDesigns(Inputs)

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    ExpDesign.method = 'sequential'
    ExpDesign.n_init_samples = 50  # 00#3*ndim

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) grid(FT) 8)user
    ExpDesign.sampling_method = 'sobol'

    # Provide the experimental design object with a hdf5 file
    # ExpDesign.hdf5_file = 'ExpDesign_AnalyticFunc.hdf5'

    # Set the sampling parameters
    ExpDesign.n_new_samples = 1
    ExpDesign.n_max_samples = 51  # 150          # sum of init + sequential
    ExpDesign.mod_LOO_threshold = 1e-16

    # ExpDesign.adapt_verbose = True
    # 1) None 2) 'equal' 3)'epsilon-decreasing' 4) 'adaptive'
    ExpDesign.tradeoff_scheme = None
    # MetaModelOpts.ExpDesign.n_replication = 5
    # -------- Exploration ------
    # 1)'Voronoi' 2)'random' 3)'latin_hypercube' 4)'LOOCV' 5)'dual annealing'
    ExpDesign.explore_method = 'random'

    # Use when 'dual annealing' chosen
    ExpDesign.max_func_itr = 1000

    # Use when 'Voronoi' or 'random' or 'latin_hypercube' chosen
    ExpDesign.n_canddidate = 1000
    ExpDesign.n_cand_groups = 4

    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'BayesActDesign' 3)'VarOptDesign' 4)'alphabetic'
    # 5)'Space-filling'
    ExpDesign.exploit_method = 'Space-filling'
    ExpDesign.exploit_method = 'BayesActDesign'
    ExpDesign.util_func = 'DKL'

    # BayesOptDesign/BayesActDesign -> when data is available
    # 1) MI (Mutual information) 2) ALC (Active learning McKay)
    # 2)DKL (Kullback-Leibler Divergence) 3)DPP (D-Posterior-percision)
    # 4)APP (A-Posterior-percision)  # ['DKL', 'BME', 'infEntropy']
    # ExpDesign.util_func = 'DKL'

    # BayesActDesign -> when data is available
    # 1) BME (Bayesian model evidence) 2) infEntropy (Information entropy)
    # 2)DKL (Kullback-Leibler Divergence)
    # ExpDesign.util_func = 'DKL'

    # VarBasedOptDesign -> when data is not available
    # 1)ALM 2)EIGF, 3)LOOCV
    # or a combination as a list
    # ExpDesign.util_func = 'EIGF'

    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality) or a combination as a list
    # ExpDesign.util_func = 'D-Opt'

    # Defining the measurement error, if it's known a priori
    obsData = pd.DataFrame(Model.observations, columns=Model.Output.names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData ** 2
    MetaModelOpts.Discrepancy = DiscrepancyOpts

    # Plot the posterior snapshots for SeqDesign
    ExpDesign.post_snapshot = False
    ExpDesign.step_snapshot = 1
    ExpDesign.max_a_post = [0] * ndim

    # For calculation of validation error for SeqDesign
    prior = np.load(f"data/Prior_{ndim}.npy")
    prior_outputs = np.load(f"data/origModelOutput_{ndim}.npy")
    likelihood = np.load(f"data/validLikelihoods_{ndim}.npy")
    ExpDesign.valid_samples = prior[:500]
    ExpDesign.valid_model_runs = {'Z': prior_outputs[:500]}

    # Run using the engine
    engine = Engine(MetaModelOpts, Model, ExpDesign)
    # engine.train_sequential()
    engine.train_normal()

    # Load the objects
    # with open(f"PCEModel_{Model.name}.pkl", "rb") as input:
    #     PCEModel = joblib.load(input)

    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    PostPCE = PostProcessing(engine)

    # Plot to check validation visually.
    PostPCE.valid_metamodel(n_samples=1)

    # Compute and print RMSE error
    PostPCE.check_accuracy(n_samples=300)

    # =====================================================
    # ========  Bayesian inference with Emulator ==========
    # =====================================================
    BayesOpts = BayesInference(engine)
    BayesOpts.emulator = True
    BayesOpts.plot_post_pred = True

    # BayesOpts.selected_indices = [0, 3, 5,  7, 9]
    # BME Bootstrap
    BayesOpts.bootstrap = True
    BayesOpts.n_bootstrap_itrs = 500
    BayesOpts.bootstrap_noise = 100

    # Bayesian cross validation
    BayesOpts.bayes_loocv = True  # TODO: test what this does

    # Select the inference method
    import emcee

    BayesOpts.inference_method = "MCMC"
    # Set the MCMC parameters passed to self.mcmc_params
    BayesOpts.mcmc_params = {
        'n_steps': 1e4,  # 5,
        'n_walkers': 30,
        'moves': emcee.moves.KDEMove(),
        'multiprocessing': False,
        'verbose': False
    }

    # ----- Define the discrepancy model -------
    obsData = pd.DataFrame(Model.observations, columns=Model.Output.names)
    BayesOpts.measurement_error = obsData

    # # -- (Option B) --
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData ** 2
    BayesOpts.Discrepancy = DiscrepancyOpts

    # -- (Option C) --
    if 0:
        DiscOutputOpts = Input()
        # # # OutputName = 'Z'
        DiscOutputOpts.add_marginals()
        DiscOutputOpts.Marginals[0].Nnme = '$\sigma^2_{\epsilon}$'
        DiscOutputOpts.Marginals[0].dist_type = 'uniform'
        DiscOutputOpts.Marginals[0].parameters = [0, 10]
        # BayesOpts.Discrepancy = {'known': DiscrepancyOpts,
        #                          'infer': Discrepancy(DiscOutputOpts)}

        BayesOpts.bias_inputs = {'Z': np.arange(0, 10, 1.).reshape(-1, 1) / 9}

        DiscOutputOpts = Input()
        # OutputName = 'lambda'
        DiscOutputOpts.add_marginals()
        DiscOutputOpts.Marginals[0].name = '$\lambda$'
        DiscOutputOpts.Marginals[0].dist_type = 'uniform'
        DiscOutputOpts.Marginals[0].parameters = [0, 1]

        # # OutputName = 'sigma_f'
        DiscOutputOpts.add_marginals()
        DiscOutputOpts.Marginals[1].Name = '$\sigma_f$'
        DiscOutputOpts.Marginals[1].dist_type = 'uniform'
        DiscOutputOpts.Marginals[1].parameters = [0, 1e-4]
        # BayesOpts.Discrepancy = Discrepancy(DiscOutputOpts)
        BayesOpts.Discrepancy = {'known': DiscrepancyOpts,
                                 'infer': Discrepancy(DiscOutputOpts)}

    # Start the calibration/inference
    Bayes_PCE = BayesOpts.create_inference()

    # Save class objects
    with open(f'Bayes_{Model.name}.pkl', 'wb') as output:
        joblib.dump(Bayes_PCE, output, 2)
