# -*- coding: utf-8 -*-
"""
Test the MetaModel class in bayesvalidrox.
Class MetaModel: 
    build_metamodel  - x
    update_metamodel
    update_pce_coeffs
    create_basis_indices --removed, just redirects
    add_InputSpace                                   -x
    univ_basis_vals
    fit
    adaptive_regression
    pca_transformation
    eval_metamodel
    create_model_error
    eval_model_error
    auto_vivification
    copy_meta_model_opts
    __select_degree
    generate_polynomials
    calculate_moments
    
"""
import numpy as np
import pytest
import sys

sys.path.append("../src/")

from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.input_space import InputSpace
from bayesvalidrox.surrogate_models.pce_gpr import PCEGPR

@pytest.fixture
def pce_gp(): 
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    samples = np.array([[0.2], [0.8]])
    MetaModel = PCEGPR(input)
    MetaModel.CollocationPoints = samples
    MetaModel.InputSpace = InputSpace(MetaModel.input_obj, MetaModel.meta_model_type)
    MetaModel.InputSpace.init_param_space(np.max(MetaModel._pce_deg))
    return MetaModel


#%% Test constructor on its own

def test_constructor() -> None:
    """
    Construct PCE without inputs
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    PCEGPR(inp)


#%% Test is_gaussian
def test_check_is_gaussian() -> None:
    """
    Check if Gaussian
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    MetaModel = PCEGPR(inp)
    MetaModel.check_is_gaussian()
    assert MetaModel.is_gaussian == True, "Expected is_gaussian to be True"

#%% Test build_metamodel

def test_build_metamodel(pce_gp) -> None:
    """
    Build PCE
    """
    mm = pce_gp
    mm.build_metamodel()

#%% Test add_InputSpace

def test_add_inputspace(pce_gp) -> None:
    """
    Add InputSpace
    """
    pce_gp.add_InputSpace()


#%% Test fit
# Faster without these
def test_fit(pce_gp) -> None:
    """
    Fit PCE
    """
    pce_gp.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})


# def test_fit_parallel(pce_gp) -> None:
#     """
#     Fit PCE in parallel
#     """
#     pce_gp.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, parallel=True)


# def test_fit_verbose(pce_gp) -> None:
#     """
#     Fit PCE verbose
#     """
#     pce_gp.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, verbose=True)

# def test_fit_parallelverbose(pce_gp) -> None:
#     """
#     Fit PCE verbose
#     """
#     pce_gp.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, parallel = True, verbose=True)


#%% Test eval_metamodel

def test_eval_metamodel(pce_gp) -> None:
    """
    Eval trained PCE 
    """
    mm = pce_gp
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    mm.eval_metamodel([[0.4]])


def test_eval_metamodel_normalboots(pce_gp) -> None:
    """
    Eval trained PCE with normal bootstrap
    """
    mm = pce_gp
    mm.bootstrap_method = 'normal'
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    mm.eval_metamodel([[0.4]])


def test_eval_metamodel_highnormalboots(pce_gp) -> None:
    """
    Eval trained PCE with higher bootstrap-itrs
    """
    mm=pce_gp
    mm.n_bootstrap_itrs = 2
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    mm.eval_metamodel([[0.4]])


def test_eval_metamodel_pca(pce_gp) -> None:
    """
    Eval trained PCE with pca
    """
    mm = pce_gp
    mm.dim_red_method = 'pca'
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.8]], {'Z': [[0.4, 0.4], [0.5, 0.6]]})
    mm.eval_metamodel([[0.4]])

#%% Test auto_vivification
def test_auto_vivification(pce_gp) -> None:
    """
    Creation of auto-vivification objects
    """
    pce_gp.auto_vivification()


#%% Test copy_meta_model_opts

def test_copy_meta_model_opts(pce_gp) -> None:
    """
    Copy the PCE with just some stats
    """
    pce_gp.copy_meta_model_opts()

#%% Test calculate_moments

def test_calculate_moments(pce_gp) -> None:
    """
    Calculate moments of a pce-surrogate
    """
    mm = pce_gp
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    mm.calculate_moments()


def test_calculate_moments_pca(pce_gp) -> None:
    """
    Calculate moments of a pce-surrogate with pca
    """
    mm = pce_gp
    mm.dim_red_method = 'pca'
    mm.fit([[0.2], [0.8]], {'Z': [[0.4, 0.4], [0.5, 0.6]]})
    mm.calculate_moments()
